# -*- coding: utf-8 -*-
"""
LeoTranspiler: Transpile a Machine Learning model to Leo.

This package provides functionalities for transpiling a Machine Learning model to Leo.
It works with scikit-learn models.
"""
from .leo_transpiler import LeoTranspiler  # noqa: F401
